/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.sql.DataSource;

import org.apache.commons.dbcp.BasicDataSource;
import org.apache.fop.apps.MimeConstants;
import org.xml.sax.ContentHandler;
import org.xml.sax.helpers.AttributesImpl;

import com.koutra.dist.proc.faucet.ReaderFaucet;
import com.koutra.dist.proc.faucet.ResultSetFaucet;
import com.koutra.dist.proc.faucet.StreamFaucet;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.pipeline.IdentityReaderToWriterPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.IdentityStreamToStreamPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.LineBasedReaderDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.LineBasedReaderToXMLDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.LineBasedReaderToXMLPipelineItem;
import com.koutra.dist.proc.pipeline.LineBasedReaderToXMLPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.ReaderDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ReaderToXMLDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ReaderToXMLPipelineItem;
import com.koutra.dist.proc.pipeline.ReaderToXMLPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.ResultSetToStreamDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ResultSetToStreamPipelineItem;
import com.koutra.dist.proc.pipeline.ResultSetToWriterDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ResultSetToWriterPipelineItem;
import com.koutra.dist.proc.pipeline.ResultSetToXMLDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.ResultSetToXMLPipelineItem;
import com.koutra.dist.proc.pipeline.StreamDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.StreamToXMLDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.StreamToXMLPipelineItem;
import com.koutra.dist.proc.pipeline.StreamToXMLPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.XMLFilterPipelineItem;
import com.koutra.dist.proc.pipeline.XMLFilterPipelineItemTemplate;
import com.koutra.dist.proc.sink.FOPSink;
import com.koutra.dist.proc.sink.FOPZipMuxSink;
import com.koutra.dist.proc.sink.LineBasedJDBCSink;
import com.koutra.dist.proc.sink.ReaderJDBCSink;
import com.koutra.dist.proc.sink.StreamJDBCSink;
import com.koutra.dist.proc.sink.StreamSink;
import com.koutra.dist.proc.sink.StreamZipMuxSink;
import com.koutra.dist.proc.sink.WriterSink;
import com.koutra.dist.proc.sink.WriterZipMuxSink;
import com.koutra.dist.proc.sink.XMLSink;
import com.koutra.dist.proc.sink.XMLZipMuxSink;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.Util;

import junit.framework.TestCase;

public class MixedContentTest extends TestCase {

	private DataSource ds;
	
    public static DataSource setupDataSource(String connectURI) {
        BasicDataSource ds = new BasicDataSource();
        ds.setDriverClassName("com.mysql.jdbc.Driver");
        ds.setUsername("root");
        ds.setPassword("z1g0u@l@");
        ds.setUrl(connectURI);
        return ds;
    }

    @Override
	protected void setUp() throws Exception {
    	ds = setupDataSource("jdbc:mysql://localhost:3306/test_db");
    	
    	Connection c = ds.getConnection();
    	c.setAutoCommit(true);
    	c.createStatement().execute("DELETE FROM TEST_SINK");
	}
	
	public void testResultSetToXml() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		ResultSetToXMLPipelineItem transformA = new ResultSetToXMLPipelineItem("2");
		XMLFilterPipelineItem transformB = new XMLFilterPipelineItem("3", "test/row2fo.xsl");
		FOPSink sink = new FOPSink("4", MimeConstants.MIME_PDF, "test/db_result.pdf");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		transformB.hookupFaucet(transformA);
		sink.hookupFaucet(transformB);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/db_result.pdf");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testResultSetToXml/db_result.pdf");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/db_result.pdf");
		f.delete();
	}
	
	public void testResultSetDemuxToXml() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		XMLFilterPipelineItemTemplate transformB =
			new XMLFilterPipelineItemTemplate("test/row2fo.xsl");
		ResultSetToXMLDemuxPipelineItem transformA =
			new ResultSetToXMLDemuxPipelineItem("2", transformB);
		FOPZipMuxSink sink = new FOPZipMuxSink(
				"4", MimeConstants.MIME_PDF, "test/db_result.zip");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		transformB.hookupFaucet(transformA);
		transformB.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/db_result.zip"));
		for (int i = 0; i < 10; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testResultSetDemuxToXml/entry" + i + ".pdf");
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/db_result.zip");
		f.delete();
	}
	
	public void testResultSetToStream() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		ResultSetToStreamPipelineItem transformA = new ResultSetToStreamPipelineItem("2") {
			@Override
			public void transformRow(ResultSet resultSet, OutputStream outputStream) {
				try {
					String value = resultSet.getString(1) + " -> " + resultSet.getString(2) + "\n";
					outputStream.write(value.getBytes());
				} catch (Exception e) {
					throw new XformationException("Unable to transform row", e);
				}
			}
		};
		StreamSink sink = new StreamSink("3", "test/db_result.txt");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		sink.hookupFaucet(transformA);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/db_result.txt");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testResultSetToStream/db_result.txt");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/db_result.txt");
		f.delete();
	}
	
	public void testResultSetDemuxToStream() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		IdentityStreamToStreamPipelineItemTemplate transformB =
			new IdentityStreamToStreamPipelineItemTemplate();
		ResultSetToStreamDemuxPipelineItem transformA =
			new ResultSetToStreamDemuxPipelineItem("2", transformB) {
			@Override
			public void transformRow(ResultSet resultSet, OutputStream outputStream) {
				try {
					String value = resultSet.getString(1) + " -> " + resultSet.getString(2) + "\n";
					outputStream.write(value.getBytes());
				} catch (Exception e) {
					throw new XformationException("Unable to transform row", e);
				}
			}
		};
		StreamZipMuxSink sink = new StreamZipMuxSink("4", "test/db_result.zip");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		transformB.hookupFaucet(transformA);
		sink.hookupFaucet(transformB);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/db_result.zip"));
		for (int i = 0; i < 10; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testResultSetDemuxToStream/entry" + i);
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/db_result.zip");
		f.delete();
	}
	
	public void testResultSetToWriter() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		ResultSetToWriterPipelineItem transformA = new ResultSetToWriterPipelineItem("2") {
			@Override
			public void transformRow(ResultSet resultSet, Writer writer) {
				try {
					String value = resultSet.getString(1) + " -> " + resultSet.getString(2) + "\n";
					writer.write(value);
				} catch (Exception e) {
					throw new XformationException("Unable to transform row", e);
				}
			}
		};
		WriterSink sink = new WriterSink("3", "test/db_result.txt", "UTF-8");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		sink.hookupFaucet(transformA);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/db_result.txt");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testResultSetToWriter/db_result.txt");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/db_result.txt");
		f.delete();
	}

	public void testResultSetDemuxToWriter() throws IOException {
		ResultSetFaucet faucet = new ResultSetFaucet("1", ds, "SELECT * FROM test_alt");
		IdentityReaderToWriterPipelineItemTemplate transformB =
			new IdentityReaderToWriterPipelineItemTemplate();
		ResultSetToWriterDemuxPipelineItem transformA =
			new ResultSetToWriterDemuxPipelineItem("2", transformB) {
			@Override
			public void transformRow(ResultSet resultSet, Writer writer) {
				try {
					String value = resultSet.getString(1) + " -> " + resultSet.getString(2) + "\n";
					writer.write(value);
				} catch (Exception e) {
					throw new XformationException("Unable to transform row", e);
				}
			}
		};
		WriterZipMuxSink sink = new WriterZipMuxSink("3", "UTF-8", "test/db_result.zip");
		
		// hook everything up
		transformA.hookupFaucet(faucet);
		transformB.hookupFaucet(transformA);
		sink.hookupFaucet(transformB);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/db_result.zip"));
		for (int i = 0; i < 10; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testResultSetDemuxToWriter/entry" + i);
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/db_result.zip");
		f.delete();
	}
	
	public void testStreamJDBCSink() throws SQLException {
		StreamFaucet faucet = new StreamFaucet("1", "test/input.txt");
		StreamJDBCSink sink = new StreamJDBCSink("2", ds) {
			private int counter = 1;
			@Override
			public void transformBuffer(Connection connection, byte[] buffer,
					int off, int len) {
				try {
					PreparedStatement stmt = connection.prepareStatement(
							"INSERT INTO TEST_SINK VALUES (0, ?, ?)");
					stmt.setString(1, String.valueOf(counter++));
					stmt.setString(2, ByteArrayUtil.getHexDump(buffer, off, len));
					stmt.execute();
				} catch (SQLException sqle) {
					throw new RuntimeException(
							"Unable to transform into the database", sqle);
				}
			}
		};
		
		// hook everything up
		sink.hookupFaucet(faucet);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the results
		Statement stmt = ds.getConnection().createStatement();
		try {
			ResultSet rs = stmt.executeQuery("SELECT * FROM TEST_SINK");
			assertTrue(rs.next());
			assertTrue(rs.getInt(2) == 1);
			assertTrue("[54,68,69,73,20,...](81)".equals(rs.getString(3)));
			assertFalse(rs.next());
			rs.close();
		} finally {
			stmt.close();
		}
	}
	
	public void testReaderJDBCSink() throws SQLException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/input.txt", "UTF-8");
		ReaderJDBCSink sink = new ReaderJDBCSink("2", ds) {
			private int counter = 1;
			@Override
			public void transformBuffer(Connection connection, char[] buffer,
					int off, int len) {
				try {
					PreparedStatement stmt = connection.prepareStatement(
							"INSERT INTO TEST_SINK VALUES (0, ?, ?)");
					stmt.setString(1, String.valueOf(counter++));
					stmt.setString(2, new String(buffer, off, len));
					stmt.execute();
				} catch (SQLException sqle) {
					throw new RuntimeException(
							"Unable to transform into the database", sqle);
				}
			}
		};
		
		// hook everything up
		sink.hookupFaucet(faucet);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the results
		Statement stmt = ds.getConnection().createStatement();
		try {
			ResultSet rs = stmt.executeQuery("SELECT * FROM TEST_SINK");
			assertTrue(rs.next());
			assertTrue(rs.getInt(2) == 1);
			String expectedValue = "This is a test.\r\n" +
				"Please test the simple stream interface.\r\n" +
				"Thank you for playing.";
			String actualValue = rs.getString(3);
			assertTrue(expectedValue.equals(actualValue));
			assertFalse(rs.next());
			rs.close();
		} finally {
			stmt.close();
		}
	}
	
	public void testLineBasedJDBCSink() throws SQLException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/input.txt", "UTF-8");
		LineBasedJDBCSink sink = new LineBasedJDBCSink("2", ds) {
			private int counter = 1;
			@Override
			public void transformLine(Connection connection, String line) {
				try {
					PreparedStatement stmt = connection.prepareStatement(
							"INSERT INTO TEST_SINK VALUES (0, ?, ?)");
					stmt.setString(1, String.valueOf(counter++));
					stmt.setString(2, line);
					stmt.execute();
				} catch (SQLException sqle) {
					throw new RuntimeException(
							"Unable to transform into the database", sqle);
				}
			}
		};
		
		// hook everything up
		sink.hookupFaucet(faucet);
		
		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();
		
		// Verify the results
		Statement stmt = ds.getConnection().createStatement();
		try {
			ResultSet rs = stmt.executeQuery("SELECT * FROM TEST_SINK");
			assertTrue(rs.next());
			assertTrue(rs.getInt(2) == 1);
			assertTrue("This is a test.".equals(rs.getString(3)));
			assertTrue(rs.next());
			assertTrue(rs.getInt(2) == 2);
			assertTrue("Please test the simple stream interface.".equals(rs.getString(3)));
			assertTrue(rs.next());
			assertTrue(rs.getInt(2) == 3);
			assertTrue("Thank you for playing.".equals(rs.getString(3)));
			assertFalse(rs.next());
			rs.close();
		} finally {
			stmt.close();
		}
	}
	
	public void testSimpleLineReaderToXML() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/simpleLineInput.txt", "UTF-8");
		LineBasedReaderToXMLPipelineItem item = new LineBasedReaderToXMLPipelineItem("2") {
			@Override
			public void transformLine(String line, ContentHandler contentHandler) {
				if (line == null)
					return;
				
				try {
					AttributesImpl attributes = new AttributesImpl();
					attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
					contentHandler.startElement("", "line", "line", attributes);
					
					String[] nameValPairs = line.split("\\|");
					for (int i = 0; i < nameValPairs.length; i++) {
						String nameValPair = nameValPairs[i];
						int equalsIndex = nameValPair.indexOf('=');
						if (equalsIndex == -1) {
							// malformed name-value pair. ignore
							continue;
						}
						
						String name = nameValPair.substring(0, equalsIndex);
						String value = nameValPair.substring(equalsIndex + 1);
						
						AttributesImpl atts = new AttributesImpl();
						atts.addAttribute("", "name", "name", "CDATA", name);
						atts.addAttribute("", "value", "value", "CDATA", value);
						
						contentHandler.startElement("", "pair", "pair", atts);
						contentHandler.endElement("", "pair", "pair");
					}
					
					contentHandler.endElement("", "line", "line");
				} catch (Exception e) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
		};
		XMLSink sink = new XMLSink("3", "test/result.xml");
		
		// hook everything up
		item.hookupFaucet(faucet);
		item.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/result.xml");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testSimpleLineReaderToXML/result.xml");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/result.xml");
		f.delete();
	}
	
	public void testReaderToXML() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/simpleLineInput.txt", "UTF-8");
		ReaderToXMLPipelineItem item = new ReaderToXMLPipelineItem("2") {
			private char[] remainingBytes = new char[0];
			@Override
			public void transformBuffer(char[] buffer, int offset, int length,
					ContentHandler contentHandler) {
				if (length != -1) {
					String oldBytes = new String(remainingBytes);
					String incBytes = new String(buffer, offset, length);
					String newBytes = oldBytes + incBytes;
					remainingBytes = newBytes.toCharArray();
				}
				
				String totalChars = new String(remainingBytes);
				if (totalChars.length() == 0)
					return;
				
				int newLineIndex = totalChars.indexOf('\n');
				String line = totalChars;
				if (newLineIndex != -1) {
					line = totalChars.substring(0, newLineIndex);
					if (line.charAt(line.length() - 1) == '\r')
						line = line.substring(0, line.length() - 1);
					remainingBytes = totalChars.substring(newLineIndex + 1).toCharArray();
				} else {
					remainingBytes = new char[0];
				}
				
				try {
					AttributesImpl attributes = new AttributesImpl();
					attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
					contentHandler.startElement("", "line", "line", attributes);
					
					String[] nameValPairs = line.split("\\|");
					for (int i = 0; i < nameValPairs.length; i++) {
						String nameValPair = nameValPairs[i];
						int equalsIndex = nameValPair.indexOf('=');
						if (equalsIndex == -1) {
							// malformed name-value pair. ignore
							continue;
						}
						
						String name = nameValPair.substring(0, equalsIndex);
						String value = nameValPair.substring(equalsIndex + 1);
						
						AttributesImpl atts = new AttributesImpl();
						atts.addAttribute("", "name", "name", "CDATA", name);
						atts.addAttribute("", "value", "value", "CDATA", value);
						
						contentHandler.startElement("", "pair", "pair", atts);
						contentHandler.endElement("", "pair", "pair");
					}
					
					contentHandler.endElement("", "line", "line");
				} catch (Exception e) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
		};
		XMLSink sink = new XMLSink("3", "test/result.xml");
		
		// hook everything up
		item.hookupFaucet(faucet);
		item.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/result.xml");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testReaderToXML/result.xml");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/result.xml");
		f.delete();
	}
	
	public void testStreamToXML() throws IOException {
		StreamFaucet faucet = new StreamFaucet("1", "test/simpleLineInput.txt");
		StreamToXMLPipelineItem item = new StreamToXMLPipelineItem("2") {
			private byte[] remainingBytes = new byte[0];
			@Override
			public void transformBuffer(byte[] buffer, int offset, int length,
					ContentHandler contentHandler) {
				try {
					if (length != -1) {
						String oldBytes = new String(remainingBytes, "UTF-8");
						String incBytes = new String(buffer, offset, length, "UTF-8");
						String newBytes = oldBytes + incBytes;
						remainingBytes = newBytes.getBytes("UTF-8");
					}
					
					String totalChars = new String(remainingBytes, "UTF-8");
					if (totalChars.length() == 0)
						return;
					
					int newLineIndex = totalChars.indexOf('\n');
					String line = totalChars;
					if (newLineIndex != -1) {
						line = totalChars.substring(0, newLineIndex);
						if (line.charAt(line.length() - 1) == '\r')
							line = line.substring(0, line.length() - 1);
						remainingBytes = totalChars.substring(newLineIndex + 1).getBytes("UTF-8");
					} else {
						remainingBytes = new byte[0];
					}
					
					try {
						AttributesImpl attributes = new AttributesImpl();
						attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
						contentHandler.startElement("", "line", "line", attributes);
						
						String[] nameValPairs = line.split("\\|");
						for (int i = 0; i < nameValPairs.length; i++) {
							String nameValPair = nameValPairs[i];
							int equalsIndex = nameValPair.indexOf('=');
							if (equalsIndex == -1) {
								// malformed name-value pair. ignore
								continue;
							}
							
							String name = nameValPair.substring(0, equalsIndex);
							String value = nameValPair.substring(equalsIndex + 1);
							
							AttributesImpl atts = new AttributesImpl();
							atts.addAttribute("", "name", "name", "CDATA", name);
							atts.addAttribute("", "value", "value", "CDATA", value);
							
							contentHandler.startElement("", "pair", "pair", atts);
							contentHandler.endElement("", "pair", "pair");
						}
						
						contentHandler.endElement("", "line", "line");
					} catch (Exception e) {
						e.printStackTrace();
						throw new RuntimeException(e);
					}
				} catch (UnsupportedEncodingException u) {
					u.printStackTrace();
					throw new RuntimeException(u);
				}
			}
		};
		XMLSink sink = new XMLSink("3", "test/result.xml");
		
		// hook everything up
		item.hookupFaucet(faucet);
		item.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Verify the result
		InputStream actualResultIS = new FileInputStream("test/result.xml");
		InputStream expectedResultIS = new FileInputStream(
				"test/validation/MixedContentTest/testStreamToXML/result.xml");
		Util.testPDFEquivalence(actualResultIS, expectedResultIS);
		actualResultIS.close();
		expectedResultIS.close();
		
		File f = new File("test/result.xml");
		f.delete();
	}
	
	public static class SimpleReaderToXMLPipelineItemTemplate
			extends ReaderToXMLPipelineItemTemplate {
		private char[] remainingBytes = new char[0];
		@Override
		public void transformBuffer(char[] buffer, int offset, int length,
				ContentHandler contentHandler) {
			if (length != -1) {
				String oldBytes = new String(remainingBytes);
				String incBytes = new String(buffer, offset, length);
				String newBytes = oldBytes + incBytes;
				remainingBytes = newBytes.toCharArray();
			}
			
			String totalChars = new String(remainingBytes);
			if (totalChars.length() == 0)
				return;
			
			int newLineIndex = totalChars.indexOf('\n');
			String line = totalChars;
			if (newLineIndex != -1) {
				line = totalChars.substring(0, newLineIndex);
				if (line.charAt(line.length() - 1) == '\r')
					line = line.substring(0, line.length() - 1);
				remainingBytes = totalChars.substring(newLineIndex + 1).toCharArray();
			} else {
				remainingBytes = new char[0];
			}
			
			try {
				AttributesImpl attributes = new AttributesImpl();
				attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
				contentHandler.startElement("", "line", "line", attributes);
				
				String[] nameValPairs = line.split("\\|");
				for (int i = 0; i < nameValPairs.length; i++) {
					String nameValPair = nameValPairs[i];
					int equalsIndex = nameValPair.indexOf('=');
					if (equalsIndex == -1) {
						// malformed name-value pair. ignore
						continue;
					}
					
					String name = nameValPair.substring(0, equalsIndex);
					String value = nameValPair.substring(equalsIndex + 1);
					
					AttributesImpl atts = new AttributesImpl();
					atts.addAttribute("", "name", "name", "CDATA", name);
					atts.addAttribute("", "value", "value", "CDATA", value);
					
					contentHandler.startElement("", "pair", "pair", atts);
					contentHandler.endElement("", "pair", "pair");
				}
				
				contentHandler.endElement("", "line", "line");
			} catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException(e);
			}
		}
	}
	
	public void testReaderToXMLTemplate() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/simpleLineInput.txt", "UTF-8");
		SimpleReaderToXMLPipelineItemTemplate template =
			new SimpleReaderToXMLPipelineItemTemplate();
		ReaderDemuxPipelineItem demux = new ReaderDemuxPipelineItem("2", template) {
			@Override
			public boolean shouldChangePipeline() {
				return false;
			}
			@Override
			public char[] transformBuffer(char[] buffer, int off, int len) {
				if (len == -1)
					return new char[0];
				
				return Arrays.copyOfRange(buffer, off, len);
			}
		};
		XMLZipMuxSink sink = new XMLZipMuxSink("3", "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		template.hookupFaucet(demux);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		ZipEntry entry = zis.getNextEntry();
		assertNotNull(entry);
		InputStream expectedIS = new FileInputStream(
				"test/validation/MixedContentTest/testReaderToXMLTemplate/entry0");
		Util.testPDFEquivalence(zis, expectedIS);
		expectedIS.close();
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/result.zip");
		f.delete();
	}
	
	public static class SimpleStreamToXMLPipelineItemTemplate
			extends StreamToXMLPipelineItemTemplate {
		private byte[] remainingBytes = new byte[0];
		@Override
		public void transformBuffer(byte[] buffer, int offset, int length,
				ContentHandler contentHandler) {
			try {
				if (length != -1) {
					String oldBytes = new String(remainingBytes, "UTF-8");
					String incBytes = new String(buffer, offset, length, "UTF-8");
					String newBytes = oldBytes + incBytes;
					remainingBytes = newBytes.getBytes("UTF-8");
				}
				
				String totalChars = new String(remainingBytes);
				if (totalChars.length() == 0)
					return;
				
				int newLineIndex = totalChars.indexOf('\n');
				String line = totalChars;
				if (newLineIndex != -1) {
					line = totalChars.substring(0, newLineIndex);
					if (line.charAt(line.length() - 1) == '\r')
						line = line.substring(0, line.length() - 1);
					remainingBytes = totalChars.substring(newLineIndex + 1).getBytes("UTF-8");
				} else {
					remainingBytes = new byte[0];
				}
				
				try {
					AttributesImpl attributes = new AttributesImpl();
					attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
					contentHandler.startElement("", "line", "line", attributes);
					
					String[] nameValPairs = line.split("\\|");
					for (int i = 0; i < nameValPairs.length; i++) {
						String nameValPair = nameValPairs[i];
						int equalsIndex = nameValPair.indexOf('=');
						if (equalsIndex == -1) {
							// malformed name-value pair. ignore
							continue;
						}
						
						String name = nameValPair.substring(0, equalsIndex);
						String value = nameValPair.substring(equalsIndex + 1);
						
						AttributesImpl atts = new AttributesImpl();
						atts.addAttribute("", "name", "name", "CDATA", name);
						atts.addAttribute("", "value", "value", "CDATA", value);
						
						contentHandler.startElement("", "pair", "pair", atts);
						contentHandler.endElement("", "pair", "pair");
					}
					
					contentHandler.endElement("", "line", "line");
				} catch (Exception e) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			} catch (UnsupportedEncodingException u) {
				u.printStackTrace();
				throw new RuntimeException(u);
			}
		}
	}

	public void testStreamToXMLTemplate() throws IOException {
		StreamFaucet faucet = new StreamFaucet("1", "test/simpleLineInput.txt");
		SimpleStreamToXMLPipelineItemTemplate template =
			new SimpleStreamToXMLPipelineItemTemplate();
		StreamDemuxPipelineItem demux = new StreamDemuxPipelineItem("2", template) {
			@Override
			public boolean shouldChangePipeline() {
				return false;
			}
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (len == -1)
					return new byte[0];
				
				return Arrays.copyOfRange(buffer, off, len);
			}
		};
		XMLZipMuxSink sink = new XMLZipMuxSink("3", "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		template.hookupFaucet(demux);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		ZipEntry entry = zis.getNextEntry();
		assertNotNull(entry);
		InputStream expectedIS = new FileInputStream(
				"test/validation/MixedContentTest/testStreamToXMLTemplate/entry0");
		Util.testPDFEquivalence(zis, expectedIS);
		expectedIS.close();
		assertNull(zis.getNextEntry());
		zis.close();
		
		File f = new File("test/result.zip");
		f.delete();
	}

	public static final class IdentityLineBasedReaderToXMLPipelineItemTemplate
		extends LineBasedReaderToXMLPipelineItemTemplate {
		@Override
		public void transformLine(String line, ContentHandler contentHandler) {
			if (line == null)
				return;
			
			try {
				AttributesImpl attributes = new AttributesImpl();
				attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
				contentHandler.startElement("", "line", "line", attributes);
				
				String[] nameValPairs = line.split("\\|");
				for (int i = 0; i < nameValPairs.length; i++) {
					String nameValPair = nameValPairs[i];
					int equalsIndex = nameValPair.indexOf('=');
					if (equalsIndex == -1) {
						// malformed name-value pair. ignore
						continue;
					}
					
					String name = nameValPair.substring(0, equalsIndex);
					String value = nameValPair.substring(equalsIndex + 1);
					
					AttributesImpl atts = new AttributesImpl();
					atts.addAttribute("", "name", "name", "CDATA", name);
					atts.addAttribute("", "value", "value", "CDATA", value);
					
					contentHandler.startElement("", "pair", "pair", atts);
					contentHandler.endElement("", "pair", "pair");
				}
				
				contentHandler.endElement("", "line", "line");
			} catch (Exception e) {
				e.printStackTrace();
				throw new RuntimeException(e);
			}
		}
	}
	
	public void testLineReaderToXMLTemplate() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet("1", "test/simpleLineInput.txt", "UTF-8");
		IdentityLineBasedReaderToXMLPipelineItemTemplate template =
			new IdentityLineBasedReaderToXMLPipelineItemTemplate();
		LineBasedReaderDemuxPipelineItem demux = new LineBasedReaderDemuxPipelineItem(
				"3", template) {
			@Override
			public boolean shouldChangePipeline() {
				return false;
			}
			
			@Override
			public String transformLine(String line) {
				if (line == null || line.length() == 0)
					return "";
				
				return line + "\r\n";
			}
		};
		XMLZipMuxSink sink = new XMLZipMuxSink("3", "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		template.hookupFaucet(demux);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		ZipEntry entry = zis.getNextEntry();
		assertNotNull(entry);
		InputStream expectedIS = new FileInputStream(
				"test/validation/MixedContentTest/testLineReaderToXMLTemplate/entry0");
		Util.testPDFEquivalence(zis, expectedIS);
		expectedIS.close();
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/result.zip");
		f.delete();
	}
	
	public void testLineReaderDemuxToFOP() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet(
				"1", "test/simpleLineInput.txt", "UTF-8");
		XMLFilterPipelineItemTemplate template =
			new XMLFilterPipelineItemTemplate("test/line2fo.xsl");
		LineBasedReaderToXMLDemuxPipelineItem demux =
			new LineBasedReaderToXMLDemuxPipelineItem("2", template) {
			private boolean shouldSwitchPipeline = false;
			@Override
			public boolean shouldChangePipeline() {
				if (shouldSwitchPipeline) {
					shouldSwitchPipeline = false;
					return true;
				}
				
				return false;
			}
			@Override
			public void transformLine(String line, ContentHandler contentHandler) {
				if (line == null || line.length() == 0)
					return;
				
				shouldSwitchPipeline = true;
				try {
					AttributesImpl attributes = new AttributesImpl();
					attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
					contentHandler.startElement("", "line", "line", attributes);
					
					String[] nameValPairs = line.split("\\|");
					for (int i = 0; i < nameValPairs.length; i++) {
						String nameValPair = nameValPairs[i];
						int equalsIndex = nameValPair.indexOf('=');
						if (equalsIndex == -1) {
							// malformed name-value pair. ignore
							continue;
						}
						
						String name = nameValPair.substring(0, equalsIndex);
						String value = nameValPair.substring(equalsIndex + 1);
						
						AttributesImpl atts = new AttributesImpl();
						atts.addAttribute("", "name", "name", "CDATA", name);
						atts.addAttribute("", "value", "value", "CDATA", value);
						
						contentHandler.startElement("", "pair", "pair", atts);
						contentHandler.endElement("", "pair", "pair");
					}
					
					contentHandler.endElement("", "line", "line");
				} catch (Exception e) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
		};
		FOPZipMuxSink sink = new FOPZipMuxSink(
				"3", MimeConstants.MIME_PDF, "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		demux.hookupSink(template);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		for (int i = 0; i < 4; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testLineReaderDemuxToFOP/entry" + i + ".pdf");
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/result.zip");
		f.delete();
	}
	
	public void testReaderDemuxToFOP() throws IOException {
		ReaderFaucet faucet = new ReaderFaucet(
				"1", "test/simpleLineInput.txt", "UTF-8");
		XMLFilterPipelineItemTemplate template =
			new XMLFilterPipelineItemTemplate("test/line2fo.xsl");
		ReaderToXMLDemuxPipelineItem demux =
			new ReaderToXMLDemuxPipelineItem("2", template) {
			private char[] remainingBytes = new char[0];
			private boolean shouldSwitchPipeline = false;
			@Override
			public boolean shouldChangePipeline() {
				if (shouldSwitchPipeline) {
					shouldSwitchPipeline = false;
					return true;
				}
				
				return false;
			}
			
			@Override
			public void transformBuffer(char[] buffer, int offset,
					int length, ContentHandler contentHandler) {
				if (length != -1) {
					String oldBytes = new String(remainingBytes);
					String incBytes = new String(buffer, offset, length);
					String newBytes = oldBytes + incBytes;
					remainingBytes = newBytes.toCharArray();
				}
				
				String totalChars = new String(remainingBytes);
				if (totalChars.length() == 0)
					return;
				
				int newLineIndex = totalChars.indexOf('\n');
				String line = totalChars;
				if (newLineIndex != -1) {
					line = totalChars.substring(0, newLineIndex);
					if (line.charAt(line.length() - 1) == '\r')
						line = line.substring(0, line.length() - 1);
					remainingBytes = totalChars.substring(newLineIndex + 1).toCharArray();
				} else {
					remainingBytes = new char[0];
				}
				
				shouldSwitchPipeline = true;
				try {
					AttributesImpl attributes = new AttributesImpl();
					attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
					contentHandler.startElement("", "line", "line", attributes);
					
					String[] nameValPairs = line.split("\\|");
					for (int i = 0; i < nameValPairs.length; i++) {
						String nameValPair = nameValPairs[i];
						int equalsIndex = nameValPair.indexOf('=');
						if (equalsIndex == -1) {
							// malformed name-value pair. ignore
							continue;
						}
						
						String name = nameValPair.substring(0, equalsIndex);
						String value = nameValPair.substring(equalsIndex + 1);
						
						AttributesImpl atts = new AttributesImpl();
						atts.addAttribute("", "name", "name", "CDATA", name);
						atts.addAttribute("", "value", "value", "CDATA", value);
						
						contentHandler.startElement("", "pair", "pair", atts);
						contentHandler.endElement("", "pair", "pair");
					}
					
					contentHandler.endElement("", "line", "line");
				} catch (Exception e) {
					e.printStackTrace();
					throw new RuntimeException(e);
				}
			}
		};
		FOPZipMuxSink sink = new FOPZipMuxSink(
				"3", MimeConstants.MIME_PDF, "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		demux.hookupSink(template);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		for (int i = 0; i < 4; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testReaderDemuxToFOP/entry" + i + ".pdf");
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/result.zip");
		f.delete();
	}
	
	public void testStreamDemuxToFOP() throws IOException {
		StreamFaucet faucet = new StreamFaucet(
				"1", "test/simpleLineInput.txt");
		XMLFilterPipelineItemTemplate template =
			new XMLFilterPipelineItemTemplate("test/line2fo.xsl");
		StreamToXMLDemuxPipelineItem demux =
			new StreamToXMLDemuxPipelineItem("2", template) {
			private byte[] remainingBytes = new byte[0];
			private boolean shouldSwitchPipeline = false;
			@Override
			public boolean shouldChangePipeline() {
				if (shouldSwitchPipeline) {
					shouldSwitchPipeline = false;
					return true;
				}
				
				return false;
			}
			
			@Override
			public void transformBuffer(byte[] buffer, int offset,
					int length, ContentHandler contentHandler) {
				try {
					if (length != -1) {
						String oldBytes = new String(remainingBytes, "UTF-8");
						String incBytes = new String(buffer, offset, length, "UTF-8");
						String newBytes = oldBytes + incBytes;
						remainingBytes = newBytes.getBytes("UTF-8");
					}
					
					String totalChars = new String(remainingBytes, "UTF-8");
					if (totalChars.length() == 0)
						return;
					
					int newLineIndex = totalChars.indexOf('\n');
					String line = totalChars;
					if (newLineIndex != -1) {
						line = totalChars.substring(0, newLineIndex);
						if (line.charAt(line.length() - 1) == '\r')
							line = line.substring(0, line.length() - 1);
						remainingBytes = totalChars.substring(newLineIndex + 1).getBytes("UTF-8");
					} else {
						remainingBytes = new byte[0];
					}
					
					shouldSwitchPipeline = true;
					try {
						AttributesImpl attributes = new AttributesImpl();
						attributes.addAttribute("", "test", "test", "CDATA", "this is a test");
						contentHandler.startElement("", "line", "line", attributes);
						
						String[] nameValPairs = line.split("\\|");
						for (int i = 0; i < nameValPairs.length; i++) {
							String nameValPair = nameValPairs[i];
							int equalsIndex = nameValPair.indexOf('=');
							if (equalsIndex == -1) {
								// malformed name-value pair. ignore
								continue;
							}
							
							String name = nameValPair.substring(0, equalsIndex);
							String value = nameValPair.substring(equalsIndex + 1);
							
							AttributesImpl atts = new AttributesImpl();
							atts.addAttribute("", "name", "name", "CDATA", name);
							atts.addAttribute("", "value", "value", "CDATA", value);
							
							contentHandler.startElement("", "pair", "pair", atts);
							contentHandler.endElement("", "pair", "pair");
						}
						
						contentHandler.endElement("", "line", "line");
					} catch (Exception e) {
						e.printStackTrace();
						throw new RuntimeException(e);
					}
				} catch (UnsupportedEncodingException u) {
					u.printStackTrace();
					throw new RuntimeException(u);
				}
			}
		};
		FOPZipMuxSink sink = new FOPZipMuxSink(
				"3", MimeConstants.MIME_PDF, "test/result.zip");
		
		// hook everything up
		demux.hookupFaucet(faucet);
		demux.hookupSink(template);
		template.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things down
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/result.zip"));
		for (int i = 0; i < 4; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MixedContentTest/testStreamDemuxToFOP/entry" + i + ".pdf");
			Util.testPDFEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File f = new File("test/result.zip");
		f.delete();
	}
	
}
