/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.model;

/**
 * Implementors of this interface represents a single item in the pipeline of
 * transformations.
 * 
 * @author Pafsanias Ftakas
 */
public interface IPipelineItem extends ISink {
	/**
	 * Implementors of the pipeline item must implement this method in order to deal
	 * with the case when they are a part of a template pipeline that takes part in
	 * a demux/mux pipeline. The idea is that this method is called on the pipeline
	 * parts that take part on the demux/mux pipeline. All such parts are supposed to
	 * carry it along to their sinks, until it reaches the mux part. There the interface
	 * <code>registerSource</code> must be called in order to notify the mux part. See
	 * the documentation of the <code>registerSource</code> method for more.
	 */
	void handleMuxSwitch();
	
	/**
	 * Implementors of the pipeline item must implement this method in order to deal
	 * with the case when they are a part of a template pipeline that takes part in
	 * a demux/mux pipeline. The idea is that this method is called on the pipeline
	 * parts that take part on the demux/mux pipeline. All such parts are supposed to
	 * carry it along to their sinks, until it reaches the mux part. There, the interface
	 * <code>registerSource</code> must be called with a null parameter. This signals
	 * that there are no more mux entries and the mux must end.
	 */
	void endMux();
	
	/**
	 * Consume the faucet. This is a specialization of the <code>consume()</code>
	 * interface of the <code>ISink</code> interface used for pipeline items. When
	 * implementing this interface, the <code>ISink</code> interface should throw an
	 * exception to catch any regression problems at source.
	 * 
	 * @param previousSink The sink that the pipeline item is connected to (i.e. the
	 * next sink down the chain).
	 */
	Object consume(ISink previousSink);
}
