/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedReader;
import java.io.Reader;
import java.sql.Connection;

import javax.sql.DataSource;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.DataSourceManager;

public abstract class LineBasedJDBCSink extends ReaderJDBCSink {
	private static final Logger logger = Logger.getLogger(LineBasedJDBCSink.class);
	
	protected BufferedReader bufferedReader;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public LineBasedJDBCSink() { }
	
	public LineBasedJDBCSink(String id, DataSource ds) {
		super(id, ds);
		bufferedReader = null;
	}

	public LineBasedJDBCSink(String id, String dataSourceName) {
		super(id, dataSourceName);
		bufferedReader = null;
	}

	@Override
	public void transformBuffer(Connection connection, char[] buffer, int off,
			int len) {
		throw new XformationException("Calling transformBuffer() method of a " +
				"LineBasedJDBCSink instance is not allowed.");
	}
	
	public abstract void transformLine(Connection connection, String line);
	
	@Override
	public void consume() {
		if (!hookedUp)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		if (faucet instanceof IPipelineItem) {
			((IPipelineItem) faucet).consume(this);
		}
		
		try {
			if (dataSourceName != null) {
				ds = DataSourceManager.getInstance().getDataSource(dataSourceName);
			}
			connection = ds.getConnection();
			connection.setAutoCommit(true);
			reader = (Reader) faucet.getSource(ContentType.CharStream);
			bufferedReader = new BufferedReader(reader);
			if (logger.isTraceEnabled())
				logger.trace("Using the reader " + bufferedReader + " in the mux sink");
			String line;
			while ((line = bufferedReader.readLine()) != null) {
				if (logger.isTraceEnabled())
					logger.trace("Read " + line + " from reader " +
							reader);
				transformLine(connection, line);
			}
		} catch (Exception ioe) {
			throw new XformationException("Unable to transform stream", ioe);
		}
	}

}
